import subprocess  # nosec

import click

from magic.shared.config import SPELLBOOK_EDITOR, SPELLBOOK_PATH


def edit_spellbook():
    print(f"Opening spellbook in {click.style(SPELLBOOK_EDITOR, fg='cyan')}...")
    subprocess.call([SPELLBOOK_EDITOR, SPELLBOOK_PATH])  # nosec



import click

from magic.shared.spellbook import read_spells

RESERVED_WORDS = ["add", "edit"]


# Line validation


def is_a_number(line):
    return line.isnumeric()


def is_yes_or_no(_line):
    line = _line.lower()
    return line in ["y", "yes", "n", "no"]


# List validation


def is_empty(_list):
    if len(_list):
        return "" in _list
    return True


def has_duplicates(_list):
    return len(_list) != len(set(_list))


def __is_each_word_available(words, spells):
    errors = []
    for word in words:
        if word in RESERVED_WORDS:
            errors.append(f"Error: '{word}' is a reserved word")
        if spells.get(word):
            errors.append(f"Error: '{word}' is already used in a spell")
    return errors


def __print_validation_errors(errors):
    error_string = "\n".join(errors) + "\n"
    print(click.style(error_string, fg="yellow"))


def magic_word_validator():
    spells = read_spells()

    def validate(line):
        words = [word.strip(" ") for word in line.split(",")]

        if is_empty(words):
            return False
        if has_duplicates(words):
            return False

        errors = __is_each_word_available(words, spells)
        if len(errors) > 0:
            __print_validation_errors(errors)
            return False

        return True

    return validate



from os import path

SPELLBOOK_SCHEMA_PATH = path.join(path.dirname(__file__), "spellbook.schema.json")
SPELLBOOK_PATH = path.join(path.expanduser("~"), ".spellbook.json")
SPELLBOOK_INDENTATION = 2
SPELLBOOK_EDITOR = "code"



import click
from click import Context, Group
from importlib_metadata import version

from magic.shared.display import EMOJI_SPARKLE

NAME = "Magic"
VERSION = version("tatuarvela-magic")
YEAR = 2021
AUTHOR = "Tatu Arvela"
VERSION_TEXT = (
    f"{EMOJI_SPARKLE} {click.style(NAME, fg='blue')} v{VERSION} © {YEAR} {AUTHOR}"
)
DESCRIPTION = "A tool for simplifying repeated command line tasks"


def get_help(self, ctx: Context):
    click_help = Group.get_help(self, ctx)
    return f"{VERSION_TEXT}\n" f"{DESCRIPTION}\n" f"\n" f"{click_help}"



import subprocess  # nosec
from datetime import datetime, timedelta
from string import Template

import click

from magic.shared.display import (
    EMOJI_FAILURE,
    EMOJI_SPARKLE,
    EMOJI_SUCCESS,
    EMOJI_TIMER,
    print_error,
)
from magic.shared.spellbook import read_spell


def cast_spell(magic_word, arguments):
    start_time = datetime.now()

    try:
        # __attempt_spell returns true if showSuccessMessage is true
        show_success_message = __attempt_spell(magic_word, arguments)
        if show_success_message:
            __print_result(start_time, success=True)

    except RuntimeError:
        __print_result(start_time, success=False)


def __check_args(argument_count, spell_args):
    if argument_count is not None and len(spell_args) < argument_count:
        raise Exception(f"Not enough arguments, {argument_count} required")
    return spell_args


def __substitute_args(text, args):
    template = Template(text)
    args_dict = {f"a{index}": arg for index, arg in enumerate(args)}
    return template.substitute(**args_dict)


def __handle_message(spell, spell_args):
    description = spell.get("description")
    show_message = spell.get("showMessage")
    if show_message is not False:
        if spell_args is not None:
            description = __substitute_args(description, spell_args)
        print(f"{EMOJI_SPARKLE} {click.style(description, fg='cyan')}")


def __parse_command(command, spell_args):
    if spell_args is not None:
        command = __substitute_args(command, spell_args)
    return command


def __attempt_spell(magic_word, arguments):
    try:
        spell = read_spell(magic_word)

        if spell:
            spell_args = __check_args(spell.get("argumentCount"), arguments)
            __handle_message(spell, spell_args)

            executable_commands = ""
            for command in spell["commands"]:
                parsed_command = __parse_command(command, spell_args)
                executable_commands = f"{executable_commands}\n{parsed_command}"

            with subprocess.Popen(executable_commands, shell=True) as process:  # nosec
                exit_code = process.wait()

                if exit_code != 0:
                    raise Exception(f"Command returned exit code {exit_code}")
                return spell.get("showSuccessMessage") is not False

        raise Exception(f"Spell not found for magic word: {magic_word}")

    except Exception as error:
        print_error(error)
        raise RuntimeError from error


def __print_result(start_time, success):
    current_time = datetime.now().strftime("%H:%M:%S")
    elapsed_time = datetime.now() - start_time
    elapsed_time = elapsed_time - timedelta(microseconds=elapsed_time.microseconds)

    result_emoji = EMOJI_SUCCESS if success else EMOJI_FAILURE
    time_message = (
        click.style(current_time, fg="green")
        if success
        else click.style(current_time, fg="red")
    )

    print(f"{result_emoji} {time_message} | {EMOJI_TIMER} {elapsed_time}")



from magic.shared.display import EMOJI_TRASH
from magic.shared.prompt import yes_or_no_prompt
from magic.shared.spellbook import delete_spell as _delete_spell
from magic.show_spell import show_spell


def delete_spell(magic_word):
    print(f"{EMOJI_TRASH} Deleting spell '{magic_word}'\n")

    show_spell(magic_word, spell_args=[], skip_arguments_provided=True)
    print("")

    confirm = yes_or_no_prompt(
        "This action cannot be undone. Are you sure you want to delete this spell, yes or NO?",
        default="no",
        color="red",
    )

    if confirm:
        _delete_spell(magic_word)



import click

from magic.shared.spellbook import read_spell


def show_spell(magic_word, spell_args, skip_arguments_provided=False):
    spell = read_spell(magic_word)
    color = "cyan"

    if not spell:
        raise Exception(f"Spell not found for magic word '{magic_word}'")

    print(f'{click.style("Description:", fg=color)} {spell["description"]}')
    print(f'{click.style("Magic words:", fg=color)} {", ".join(spell["magicWords"])}')
    print(click.style("Commands:", fg=color))
    for command in spell["commands"]:
        print(f"  {command}")

    argument_count = spell.get("argumentCount")
    if argument_count is None:
        print(f'{click.style("Arguments required:", fg=color)} None')
    else:
        print(f'{click.style("Arguments required:", fg=color)} {argument_count}')

    if skip_arguments_provided:
        return

    print(f'{click.style("Arguments provided:", fg=color)} {len(spell_args)}')
    for idx, arg in enumerate(spell_args):
        print(f"  $a{idx}: {arg}")



import sys

import click

from magic.shared.display import clear_last_line
from magic.shared.validation import is_yes_or_no


def prompt(
    message,
    color,
    validate=None,
    default=None,
    required=False,
):
    print(click.style(message, fg=color))

    return __prompt_input(validate, default, required)


def multiline_prompt(
    message, color, validate=None, default=None, required=False
):
    print(click.style(message, fg=color))

    lines = []
    while True:
        line = __prompt_input(validate, default)
        if line:
            lines.append(line)
        elif not required or len(lines) > 0:
            clear_last_line()
            break
    return lines


def yes_or_no_prompt(message, color, default=None, required=False):
    print(click.style(message, fg=color))

    return __yes_or_no(__prompt_input(is_yes_or_no, default, required))


def __yes_or_no(value):
    if value in ("y", "yes"):
        return True
    return False


def __prompt_input(validate=None, default=None, required=False):
    input_message = "> "

    response = None
    while response is None:
        try:
            response = input(click.style(input_message, fg="cyan"))
        except KeyboardInterrupt:
            sys.exit()

        if default is not None:
            if response == "":
                clear_last_line()
                print(click.style(f"{input_message}{default}", fg="cyan"))
                response = default

        if required and response == "":
            response = None
            clear_last_line()

        if validate is not None:
            if validate(response) is False:
                response = None
                clear_last_line()

    return response



import re

from magic.shared.display import EMOJI_WIZARD
from magic.shared.prompt import multiline_prompt, prompt, yes_or_no_prompt
from magic.shared.spellbook import create_spell
from magic.shared.validation import magic_word_validator

PROMPT_COLOR = "cyan"


def add_spell():
    print(
        f"{EMOJI_WIZARD} This wizard will add a new spell to your .spellbook.json file.\n"
        f"Press ^C at any time to quit.\n"
    )

    commands = multiline_prompt(
        "Enter commands to be run in the spell, separated by line breaks.\n"
        "You may use $a0, $a1, etc. to provide arguments.\n"
        "Leave the line empty to continue.",
        required=True,
        color=PROMPT_COLOR,
    )
    print("")

    description = prompt(
        "Enter a short description.\n"
        "You may use arguments ($a0, $a1, etc.) that will be filled for the casting message.",
        required=True,
        color=PROMPT_COLOR,
    )
    print("")

    magic_words = [
        word.strip(" ")
        for word in prompt(
            "Enter magic words, separated by commas.\n"
            "Any of them can be used for calling the spell: 'magic <magic_word>'.",
            validate=magic_word_validator(),
            color=PROMPT_COLOR,
        ).split(",")
    ]
    print("")

    show_message = yes_or_no_prompt(
        "Do you want to show a message when casting the spell, YES or no?",
        default="yes",
        color=PROMPT_COLOR,
    )
    print("")

    show_success_message = yes_or_no_prompt(
        "Do you want to show a message on completion "
        "(containing the starting time and duration of the spell), YES or no?",
        default="yes",
        color=PROMPT_COLOR,
    )

    spell = {
        "description": description,
        "magicWords": magic_words,
        "commands": commands,
        "argumentCount": __count_arguments(description, commands),
        "showMessage": show_message,
        "showSuccessMessage": show_success_message,
    }

    create_spell(spell)


def __count_arguments(description, commands):
    arg_matcher = "\\$a[0-9]+"

    combined_strings = f"{description} {' '.join(commands)}"
    args = re.findall(arg_matcher, combined_strings)
    for index, arg in enumerate(args):
        args[index] = int(arg.replace("$a", ""))
    args.sort()

    if len(args) == 0:
        return 0
    return args[-1] + 1



import json
import os

from fastjsonschema import validate

from magic.shared.config import (
    SPELLBOOK_INDENTATION,
    SPELLBOOK_PATH,
    SPELLBOOK_SCHEMA_PATH,
)

DEFAULT_SPELL = {
    "description": "Example echo spell with arguments '$a0' and '$a1'",
    "magicWords": ["e", "example"],
    "commands": ["echo $a0", "echo $a1"],
    "argumentCount": 2,
}


def __create_spellbook():
    with open(SPELLBOOK_PATH, "x", encoding="utf-8") as file:
        json.dump([DEFAULT_SPELL], file, indent=SPELLBOOK_INDENTATION)


def __validate_spellbook(spellbook_contents):
    try:
        with open(SPELLBOOK_SCHEMA_PATH, "r", encoding="utf-8") as file:
            schema = json.load(file)
            validate(schema, spellbook_contents)
    except Exception as error:
        raise Exception(f"Spellbook is invalid: {error}") from error


def __open_spellbook():
    if not os.path.exists(SPELLBOOK_PATH):
        __create_spellbook()

    with open(SPELLBOOK_PATH, "r", encoding="utf-8") as file:
        spellbook = json.load(file)
        __validate_spellbook(spellbook)
        return spellbook


def create_spell(spell):
    with open(SPELLBOOK_PATH, "r+", encoding="utf-8") as file:
        spellbook = json.load(file)  # spells are already validated in add_spell()
        spellbook.append(spell)
        file.seek(0)
        json.dump(spellbook, file, indent=SPELLBOOK_INDENTATION)
        file.truncate()


def read_spells():
    spellbook = __open_spellbook()
    spells = {}
    for entry in spellbook:
        for magic_word in entry["magicWords"]:
            if spells.get(magic_word):
                raise Exception(f"Spellbook has duplicated magic word: {magic_word}")
            spells[magic_word] = entry
    return spells


def read_spell(magic_word):
    spells = read_spells()
    return spells.get(magic_word)


def delete_spell(magic_word):
    with open(SPELLBOOK_PATH, "r+", encoding="utf-8") as file:

        def magic_word_filter(spell):
            if magic_word in spell["magicWords"]:
                return False

            return True

        spellbook = json.load(file)  # spell validity does not matter here
        spellbook = list(filter(magic_word_filter, spellbook))
        file.seek(0)
        json.dump(spellbook, file, indent=SPELLBOOK_INDENTATION)
        file.truncate()



import sys

import click

# Emoji spacing in terminal is unpredictable
# These extra spaces appear to fix issues
EMOJI_FAILURE = "\u274c\ufeff"
EMOJI_FIRE = "\U0001f525"
EMOJI_SPARKLE = "\u2728\ufeff"
EMOJI_SUCCESS = "\u2705\ufeff"
EMOJI_TIMER = "\u23f1\u0020"
EMOJI_TRASH = "\U0001f5d1\u0020"
EMOJI_WIZARD = "\U0001f9d9"


RESET_COLOR = "\u001b[0m"


def print_error(error):
    print(click.style(f'{EMOJI_FIRE} Error: {error}', fg='red'))


def clear_last_line():
    sys.stdout.write("\033[F")  # back to previous line
    sys.stdout.write("\033[K")  # clear line



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['magic', 'magic.shared']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.0,<9.0.0',
 'fastjsonschema>=2.15.0,<3.0.0',
 'importlib-metadata>=4.6.3,<5.0.0']

entry_points = \
{'console_scripts': ['magic = magic:main']}

setup_kwargs = {
    'name': 'tatuarvela-magic',
    'version': '3.1.3',
    'description': 'A tool for simplifying repeated command line tasks',
    'long_description': '# ![Magic icon](./icon.png?raw=true "Magic icon") Magic\n\n[![Repository](https://img.shields.io/badge/repository-gray.svg?logo=github)](https://github.com/TatuArvela/magic)\n[![GitHub issues](https://img.shields.io/github/issues/TatuArvela/magic)](https://github.com/TatuArvela/magic/issues)\n[![Pipeline status](https://github.com/TatuArvela/magic/actions/workflows/verify.yml/badge.svg?event=push)](https://github.com/TatuArvela/magic/actions/workflows/verify.yml)\n[![PyPI](https://img.shields.io/pypi/v/tatuarvela-magic)](https://pypi.org/project/tatuarvela-magic/)\n[![License](https://img.shields.io/github/license/TatuArvela/magic)](https://github.com/TatuArvela/magic/blob/master/LICENSE)\n[![Created at Nitor](https://img.shields.io/badge/created%20at-Nitor-informational.svg)](https://nitor.com/)\n\nMagic is a tool for turning repeated command line tasks and long, hard to\nremember commands into quickly callable simple commands, **spells**.\n\nSpells provide a simple and managed alternative to manually created aliases or\nscripts.\n\n## 🏃 Quick tour\n\n1. To add spells, run the built-in **wizard** `magic add`\n2. Spells are written into the **spellbook** file (`~/.spellbook.json`)\n3. Each spell is available as a command under `magic`, which can be listed\n   with `magic --help`\n4. A spell can have one or several command names, which are called **magic\n   words**  \n   e.g. `magic build-app` and `magic ba`\n5. Spells can have **arguments** passed to them  \n   e.g. `magic say abra kadabra`\n\n## 💻 Installation\n\nMagic is designed for macOS and common Linux distributions using Bash or Zsh.\nMagic also works on Windows Subsystem for Linux.\n\nMagic requires Python 3.7 or above, and can be installed using pip:\n\n```console\npython3 -m pip install tatuarvela-magic\n```\n\n## 🪄 Usage\n\n```console\n$ magic\n✨ Magic © 2022 Tatu Arvela\nA tool for simplifying repeated command line tasks\n\nUsage: magic [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  -h, --help  Show this message and exit.\n\nCommands:\n  add      Add spell to spellbook\n  e        Example echo spell with arguments \'$a0\' and \'$a1\'\n  edit     Open spellbook in editor\n  example  Example echo spell with arguments \'$a0\' and \'$a1\'\n```\n\nEditing the spellbook is currently done with an external editor (**Visual Studio\nCode** by default).\n\n## 📚 Documentation\n\n### Spell options\n\nSpell options can be listed with the `-h` or `--help` option.\n\n```console\n$ magic example --help\nUsage: magic example [OPTIONS]\n\n  Example echo spell with arguments \'$a0\' and \'$a1\'\n\nOptions:\n  -d, --delete  Delete this spell.\n  -s, --show    Show details of this spell.\n  -h, --help    Show this message and exit.\n```\n\n`-d` or `--delete` option can be used to delete a spell.\n\n`-s` or `--show` option can be used to show the details of a spell.\n\nOther options are interpreted as arguments for spells.\n\n### Spell arguments\n\nSpells can have an array of arguments, which are populated according to their\nindex, starting from 0.\n\nExample:\n\n```json\n{\n  "description": "Example echo spell with arguments \'$a0\' and \'$a1\'",\n  "magicWords": [\n    "e",\n    "example"\n  ],\n  "commands": [\n    "echo $a0",\n    "echo $a1"\n  ],\n  "argumentCount": 2\n}\n```\n\n```console\n$ magic example cat dog\n✨ Example echo spell with arguments \'cat\' and \'dog\'\ncat\ndog\n✅ 12:30:43 | ⏱ 0:00:00\n```\n\nThe arguments can be used in the spell description and commands.\n\nThe description arguments are replaced when displaying the spell message.\n\n`argumentCount` property is automatically inferred from the description and\ncommands.\n\nExcessive usage of arguments is considered to be an anti-pattern, it is\nrecommended to create separate spells instead.\n\n#### Advanced usage: Empty arguments\n\nArgument are handled as an ordered array. If necessary, it is possible to make\nan argument an empty string: `\'\'`.\n\n### Messages\n\n#### Message\n\nMagic can print the description of a spell, filled with the provided arguments\n\n```console\n✨ Example echo spell with arguments \'cat\' and \'dog\'\n```\n\n`showMessage` property defaults to `true`.\n\n#### Success message\n\nMagic can show a success message which reports the starting time and duration of\na spell. This may be useful for longer operations.\n\n```console\n✅ 23:46:43 | ⏱ 0:00:00\n```\n\n`showSuccessMessage` property defaults to `true`.\n\n## ⚙️ Development\n\nDevelopment instructions and notes can be found\nin [DEVELOPMENT.md](./DEVELOPMENT.md)\n',
    'author': 'Tatu Arvela',
    'author_email': 'tatu.arvela@nitor.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/TatuArvela/magic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import click

from magic.cast_spell import cast_spell
from magic.delete_spell import delete_spell
from magic.shared.spellbook import read_spells
from magic.show_spell import show_spell


def generate_commands(group):
    for magic_word, spell in read_spells().items():
        description = click.style(spell.get("description"), fg='blue')
        group.command(
            name=magic_word,
            help=description,
            short_help=description,
            context_settings=dict(allow_extra_args=True, ignore_unknown_options=True),
        )(__create_command(magic_word))


def __create_command(magic_word):
    @click.pass_context
    @click.option("-d", "--delete", is_flag=True, help="Delete this spell.")
    @click.option("-s", "--show", is_flag=True, help="Show details of this spell.")
    def command(ctx, delete, show):
        if delete:
            return delete_spell(magic_word)
        if show:
            return show_spell(magic_word=magic_word, spell_args=ctx.args)
        return cast_spell(magic_word=magic_word, arguments=ctx.args)

    return command



import click


class MagicGroup(click.Group):
    def __init__(self, **attrs):
        super().__init__(**attrs)
        from magic.generate_commands import generate_commands

        generate_commands(self)

    def get_help(self, ctx):
        from magic.get_help import get_help

        return get_help(self, ctx)


@click.group(
    cls=MagicGroup,
    context_settings=dict(allow_extra_args=True, help_option_names=["-h", "--help"]),
)
def main():
    pass


@main.command(name="add")
def __add():
    """Add spell to spellbook"""
    from magic.add_spell import add_spell

    return add_spell()


@main.command(name="edit")
def __edit():
    """Open spellbook in editor"""
    from magic.edit_spellbook import edit_spellbook

    return edit_spellbook()
